// NVSPort2.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

// Requires that the following registry settings are set to 1:
//[HKEY_LOCAL_MACHINE\Juniper\Expansion\GPIO_16]
//    "Function"=dword:1                          ; 0 - not in use, 1 = COM4_RXD (pins 2 through 5 must be function 0)
//[HKEY_LOCAL_MACHINE\Juniper\Expansion\GPIO_17]
//    "Function"=dword:1                          ; 0 - not in use, 1 = COM4_TXD (pins 2 through 5 must be function 0)


#define GPS_COM_PORT        L"COM4:"

#define BINR_ERROR 1
#define MESSAGE_MAX         1024 // Max message size as per BINR specification, version 1.3, section 5.1 paragraph 5.

HANDLE g_hNVSPort;

typedef struct _Packet70Version
{
    BYTE numChan;                           // Number of channels
    BYTE EquipmentSoftwareVersionID[21];    // Equipment and Software Version Identifier
    DWORD SerialNumber;                     // Serial Number (device ID)
    BYTE bReserved1[21];                    // Reserved 1
    DWORD dwReserved2;                      // Reserved 2
    BYTE bReserved3[21];                    // Reserved 3
    DWORD dwReserved4;                      // Reserved 4
} PACKET70_VERSION;

typedef struct _Packet43TestResults
{
    BYTE responseType;                      // Response Type
    BYTE TestResults;                       // bit 0-3 Reserved, bit 4 GPS test, 5 GLONASS test, 6-7 Antenna: 0 connected, 1 not connected, 3 shorted
    DWORD dwReserved;                       // Reserved
} PACKET43_TESTRESULTS;

unsigned short g_Table_CRC[256] =
{
    0x0000,0x1021,0x2042,0x3063,0x4084,0x50A5,0x60C6,0x70E7,
    0x8108,0x9129,0xA14A,0xB16B,0xC18C,0xD1AD,0xE1CE,0xF1EF,
    0x1231,0x0210,0x3273,0x2252,0x52B5,0x4294,0x72F7,0x62D6,
    0x9339,0x8318,0xB37B,0xA35A,0xD3BD,0xC39C,0xF3FF,0xE3DE,
    0x2462,0x3443,0x0420,0x1401,0x64E6,0x74C7,0x44A4,0x5485,
    0xA56A,0xB54B,0x8528,0x9509,0xE5EE,0xF5CF,0xC5AC,0xD58D,
    0x3653,0x2672,0x1611,0x0630,0x76D7,0x66F6,0x5695,0x46B4,
    0xB75B,0xA77A,0x9719,0x8738,0xF7DF,0xE7FE,0xD79D,0xC7BC,
    0x48C4,0x58E5,0x6886,0x78A7,0x0840,0x1861,0x2802,0x3823,
    0xC9CC,0xD9ED,0xE98E,0xF9AF,0x8948,0x9969,0xA90A,0xB92B,
    0x5AF5,0x4AD4,0x7AB7,0x6A96,0x1A71,0x0A50,0x3A33,0x2A12,
    0xDBFD,0xCBDC,0xFBBF,0xEB9E,0x9B79,0x8B58,0xBB3B,0xAB1A,
    0x6CA6,0x7C87,0x4CE4,0x5CC5,0x2C22,0x3C03,0x0C60,0x1C41,
    0xEDAE,0xFD8F,0xCDEC,0xDDCD,0xAD2A,0xBD0B,0x8D68,0x9D49,
    0x7E97,0x6EB6,0x5ED5,0x4EF4,0x3E13,0x2E32,0x1E51,0x0E70,
    0xFF9F,0xEFBE,0xDFDD,0xCFFC,0xBF1B,0xAF3A,0x9F59,0x8F78,
    0x9188,0x81A9,0xB1CA,0xA1EB,0xD10C,0xC12D,0xF14E,0xE16F,
    0x1080,0x00A1,0x30C2,0x20E3,0x5004,0x4025,0x7046,0x6067,
    0x83B9,0x9398,0xA3FB,0xB3DA,0xC33D,0xD31C,0xE37F,0xF35E,
    0x02B1,0x1290,0x22F3,0x32D2,0x4235,0x5214,0x6277,0x7256,
    0xB5EA,0xA5CB,0x95A8,0x8589,0xF56E,0xE54F,0xD52C,0xC50D,
    0x34E2,0x24C3,0x14A0,0x0481,0x7466,0x6447,0x5424,0x4405,
    0xA7DB,0xB7FA,0x8799,0x97B8,0xE75F,0xF77E,0xC71D,0xD73C,
    0x26D3,0x36F2,0x0691,0x16B0,0x6657,0x7676,0x4615,0x5634,
    0xD94C,0xC96D,0xF90E,0xE92F,0x99C8,0x89E9,0xB98A,0xA9AB,
    0x5844,0x4865,0x7806,0x6827,0x18C0,0x08E1,0x3882,0x28A3,
    0xCB7D,0xDB5C,0xEB3F,0xFB1E,0x8BF9,0x9BD8,0xABBB,0xBB9A,
    0x4A75,0x5A54,0x6A37,0x7A16,0x0AF1,0x1AD0,0x2AB3,0x3A92,
    0xFD2E,0xED0F,0xDD6C,0xCD4D,0xBDAA,0xAD8B,0x9DE8,0x8DC9,
    0x7C26,0x6C07,0x5C64,0x4C45,0x3CA2,0x2C83,0x1CE0,0x0CC1,
    0xEF1F,0xFF3E,0xCF5D,0xDF7C,0xAF9B,0xBFBA,0x8FD9,0x9FF8,
    0x6E17,0x7E36,0x4E55,0x5E74,0x2E93,0x3EB2,0x0ED1,0x1EF0
};

void  CalculateChecksum(WORD *wChecksum, BYTE *pMessage, DWORD dwLength)
{
    DWORD index;
    BYTE *bPtr;


    *wChecksum = 0;
    bPtr = pMessage;

    for (index = 0; index < dwLength; index++)
    {
        // Checksum computation iteration
        *wChecksum = (*wChecksum<<8)^g_Table_CRC[(*wChecksum>>8)^*bPtr++];
    }
}



DWORD NVSOpenPort()
{
    DCB settings;

    //write data to NVS receiver
    g_hNVSPort = CreateFile(GPS_COM_PORT, GENERIC_WRITE | GENERIC_READ, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if(INVALID_HANDLE_VALUE != g_hNVSPort)
    {
        if(GetCommState(g_hNVSPort, &settings))
        {
            settings.BaudRate = CBR_115200;
            settings.DCBlength = sizeof(DCB);
            settings.fBinary = TRUE;
            settings.fParity = FALSE;
            settings.fOutxCtsFlow = FALSE;
            settings.fOutxDsrFlow = FALSE;
            settings.fDtrControl = DTR_CONTROL_DISABLE;
//            settings.fDtrControl = DTR_CONTROL_ENABLE;
            settings.fDsrSensitivity = FALSE;
            settings.fOutX = FALSE;
            settings.fInX = FALSE;
            settings.fErrorChar = FALSE;
            settings.fRtsControl = RTS_CONTROL_DISABLE;
//            settings.fRtsControl = RTS_CONTROL_ENABLE;
            settings.ByteSize = 8;
            settings.Parity = ODDPARITY;  // MANDATORY FOR BINR port
            settings.StopBits = ONESTOPBIT;
            if(!SetCommState(g_hNVSPort, &settings))
            {
                RETAILMSG(1, (TEXT("NVSOpenPort: SetCommState Failed\r\n")));
                CloseHandle(g_hNVSPort);
                return ERROR_GEN_FAILURE;
            }
        }
        else
        {
            RETAILMSG(1, (TEXT("NVSOpenPort: GetCommState Failed\r\n")));
            CloseHandle(g_hNVSPort);
            return ERROR_GEN_FAILURE;
        }
        
        //The receiver isn't ready immediately after opening the port
        Sleep(500);
    }
    return ERROR_SUCCESS;
}


void NVSClosePort()
{
    DCB settings;


    RETAILMSG(1, (TEXT("+NVSClosePort\r\n")));
    if (INVALID_HANDLE_VALUE != g_hNVSPort)
    {
        if(GetCommState(g_hNVSPort, &settings))
        {
            settings.fDtrControl = DTR_CONTROL_DISABLE;
            settings.fRtsControl = RTS_CONTROL_DISABLE;
            if(!SetCommState(g_hNVSPort, &settings))
            {
                RETAILMSG(1, (TEXT("NVSClosePort: SetCommState Failed\r\n")));
            }
        }
        CloseHandle(g_hNVSPort);
        g_hNVSPort = INVALID_HANDLE_VALUE;
    }
}
#if 1
#define WRITE_OK                        0
#define WRITE_ERROR_FIRST_BYTE          101
#define WRITE_ERROR_ID_BYTE             102
#define WRITE_ERROR_MSG_BYTE            103
#define WRITE_ERROR_CHECKSUM_DLE        104
#define WRITE_ERROR_CHECKSUM_CODE       105
#define WRITE_ERROR_CHECKSUM_BYTE1      106
#define WRITE_ERROR_CHECKSUM_BYTE2      107
#define WRITE_ERROR_LAST_DLE            108
#define WRITE_ERROR_ETX                 109
#define READ_OK                         0
#define READ_ERROR_BAD_PARAMETER        201
#define READ_ERROR_LENGTH_TOO_SHORT     202
#define READ_ERROR_CHECKSUM_MISMATCH    203
#define READ_ERROR_NO_CHECKSUM          204
#define READ_ERROR_NO_MESSAGE_BYTE      205
#define READ_ERROR_NO_SECOND_BYTE       206
#define READ_ERROR_NO_BYTE              207

const BYTE cbDLE = 0x10;
const BYTE cbFF  = 0xFF;
const BYTE cbCRC = 0xFF;
const BYTE cbETX = 0x03;

// Write the NVS BINR message
DWORD BINRWrite(BYTE *pMessage, DWORD length, BOOL bUsingChecksum)
{
    DWORD count;
    BYTE *bPtr;
    BYTE  bTemp;
    DWORD written;
    WORD  wChecksum;


    // Write the beginning of the message
    bTemp = cbDLE;

    if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
    {
        RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
        return WRITE_ERROR_FIRST_BYTE;
    }

    bTemp = *pMessage;

    if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
    {
        RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
        return WRITE_ERROR_ID_BYTE;
    }

    // Calculate the checksum
    CalculateChecksum(&wChecksum, pMessage, length);

    // Write the data portion (if there is one)
    if (length > 1)
    {
        bPtr = pMessage + 1;
        for(count = 0; count < length; count++)
        {
            bTemp = *bPtr++;
            if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) ||  (1 != written))
            {
                RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
                return WRITE_ERROR_MSG_BYTE;
            }
        }
    }
    if (TRUE == bUsingChecksum)
    {
        // Write the second DLE for the checksum
        bTemp = cbDLE;

        if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
        {
            RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
            return WRITE_ERROR_CHECKSUM_DLE;
        }

        // Write the "beginning of checksum" code for the checksum
        bTemp = cbFF;

        if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
        {
            RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
            return WRITE_ERROR_CHECKSUM_CODE;
        }

        // Write the checksum, endian free 
        bTemp = (wChecksum >> 8);
        
        if ((0 == WriteFile(g_hNVSPort, &bTemp, 1, &written, NULL)) ||  (1 != written))
        {
            RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
            return WRITE_ERROR_CHECKSUM_BYTE1;
        }
        bTemp = (BYTE)wChecksum;

        if ((0 == WriteFile(g_hNVSPort, &bTemp, 1, &written, NULL)) ||  (1 != written))
        {
            RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
            return WRITE_ERROR_CHECKSUM_BYTE2;
        }
    }

    bTemp = cbDLE;

    if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
    {
        RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
        return WRITE_ERROR_LAST_DLE;
    }

    // Write the "End-of-message" indicator
    bTemp = cbETX;

    if ((0 == WriteFile(g_hNVSPort, (LPVOID)&bTemp, 1, &written, NULL)) || (1 != written))
    {
        RETAILMSG(BINR_ERROR, (TEXT("BINRWrite error, %d, wrote %d\r\n"),GetLastError(), written));
        return WRITE_ERROR_ETX;
    }

    return WRITE_OK;
}


DWORD BINRRead(BYTE *pMessage, DWORD *dwLength)
{
    //DLE   0x10    1
    //ID    varies  1
    //Data  (variable length)
    //Next three may not be present
    //DLE   0x10    1
    //CRC   0xFF    1
    //CS    varies  2
    //DLE   0x10    1
    //ETX   0x03    1
    

     //Plan:
     //  While we're not "DONE":
     //       READ a byte.
     //       IF DLE, 
     //       THEN READ a byte.
     //           IF ETX, mark as "DONE".  (Means we have to prefill values in case of error)
     //                   IF ID still 0, flag "ERROR" too.
     //           IF CRC, get next RxCS and call CRC calculator on message buffer.  
     //                   IF rxCS != CalcCS flag "DONE" and "ERROR".
     //           ELSE we got an ID byte.
     //                   FILL in ID
     //                   READ a byte.  
     //                   WHile NOT DLE, 
     //                        put byte in message buffer, 
     //                        increment length
     //                        READ a byte.
     //                   Done with this loop
     //           DONE with IF
     //        ELSE
     //           We're out of sync.  Just loop around.
     //        DONE with IF
     //   DONE with this outer loop.

    BOOL  bDone = FALSE;
    DWORD dwRetVal = READ_OK;
    DWORD dwBytesRead = 0;
    BYTE  ucByte;
    WORD  wChecksum;
    WORD  wCalcChecksum;
    DWORD dwRead;
    BYTE *pMsgPtr;
    DWORD dwByteCount;

    if ((NULL == pMessage) || (MESSAGE_MAX < *dwLength))
    {
        dwRetVal = READ_ERROR_BAD_PARAMETER;
        bDone = TRUE;
    }
    else 
    {
        if ((TRUE == ReadFile(g_hNVSPort, &ucByte, 1, &dwRead, NULL)) && (1 == dwRead))
        {
            RETAILMSG(1, (TEXT("BINRRead: Read DLE byte: 0x%x\r\n"), ucByte));
            // INTENTIONALLY BLANK
        }
        else
        {
            // NOT READING ANY BYTES.  Got a problem
            RETAILMSG(1, (TEXT("BINRRead: Failed to read first byte in function\r\n")));
            bDone = TRUE;
            dwRetVal = READ_ERROR_NO_BYTE;
        }
    }

    while (!bDone)
    {
        if (cbDLE == ucByte)
        {
            if ((TRUE == ReadFile(g_hNVSPort, &ucByte, 1, &dwRead, NULL)) && (1 == dwRead))
            {
                if (cbETX == ucByte)
                {
                    RETAILMSG(1, (TEXT("BINRRead: Read byte ETX\r\n")));
                    // HANDLE ANY END CASES described above
                    bDone = TRUE;
                    *dwLength = dwByteCount;
                }
                else if (cbCRC == ucByte)
                {
                    RETAILMSG(1, (TEXT("BINRRead: Read byte CRC\r\n")));
                    if ((TRUE == ReadFile(g_hNVSPort, &wChecksum, 2, &dwRead, NULL)) && (2 == dwRead))
                    {
                        RETAILMSG(1, (TEXT("BINRRead: Read word Checksum: 0x%x\r\n"), wChecksum));
                        CalculateChecksum(&wCalcChecksum, pMessage, dwByteCount);
                        if (wChecksum != wCalcChecksum)
                        {
                            // ERROR CASE!
                            bDone = TRUE;
                            dwRetVal = READ_ERROR_CHECKSUM_MISMATCH;
                        }
                    }
                    else
                    {
                        // ERROR reading Checksum value
                        RETAILMSG(1, (TEXT("BINRRead: Failed to read the checksum value\r\n")));
                        bDone = TRUE;
                        dwRetVal = READ_ERROR_NO_CHECKSUM;
                    }
                }
                else // assume it's a message ID byte, or first byte of a message
                {
                    RETAILMSG(1, (TEXT("BINRRead: Read byte ID: 0x%x\r\n"), ucByte));
                    dwByteCount = 0;
                    pMsgPtr = pMessage;

                    if ((TRUE == ReadFile(g_hNVSPort, &ucByte, 1, &dwRead, NULL)) && (1 == dwRead))
                    {
                        RETAILMSG(1, (TEXT("BINRRead: Read byte[%d]=0x%x"),dwByteCount, ucByte));
                        if ((ucByte >= 32) && (ucByte < 127))
                        {
                            RETAILMSG(1, (TEXT(",'%c'"),ucByte));
                        }
                        RETAILMSG(1, (TEXT("\r\n")));
                        while (cbDLE != ucByte)
                        {
                            if (dwByteCount <= *dwLength)
                            {
                                *pMsgPtr++ = ucByte;
                                dwByteCount++;
                            }
                            else
                            {
                                dwRetVal = READ_ERROR_LENGTH_TOO_SHORT;
                            }
                                
                            if ((TRUE == ReadFile(g_hNVSPort, &ucByte, 1, &dwRead, NULL)) && (1 == dwRead))
                            {
                                RETAILMSG(1, (TEXT("BINRRead: Read byte[%d]=0x%x"),dwByteCount, ucByte));
                                if ((ucByte >= 32) && (ucByte < 127))
                                {
                                    RETAILMSG(1, (TEXT(",'%c'"),ucByte));
                                }
                                RETAILMSG(1, (TEXT("\r\n")));
                                // THIS SPACE INTENTIONALLY LEFT BLANK
                            }
                            else
                            {
                                // ERROR READING MESSAGE BYTE
                                RETAILMSG(1, (TEXT("BINRRead: Failed to read message byte\r\n")));
                                bDone = TRUE;
                                dwRetVal = READ_ERROR_NO_MESSAGE_BYTE;
                                break;
                            }
                        }
                    }
                    else
                    {
                        // ERROR READING FIRST MESSAGE BYTE
                        RETAILMSG(1, (TEXT("BINRRead: Failed to read the ID byte\r\n")));
                        bDone = TRUE;
                        dwRetVal = READ_ERROR_NO_MESSAGE_BYTE;
                    }
                }
            }
            else
            {
                // ERROR READING BYTE AFTER DLE
                RETAILMSG(1, (TEXT("BINRRead: Failed to read second byte after DLE\r\n")));
                bDone = TRUE;
                dwRetVal = READ_ERROR_NO_SECOND_BYTE;
                break;
            }
        }
        else
        {
            // OUT OF SYNC  
            RETAILMSG(1, (TEXT("BINRRead: OUT OF SYNC: [0x%x]\r\n"), ucByte));
        }
    }

    return dwRetVal;
}  


#endif
int _tmain(int argc, _TCHAR* argv[])
{
    BYTE message[MESSAGE_MAX];
    DWORD dwRetVal;
    DWORD dwMsgLen;

    PACKET70_VERSION vers;
    PACKET43_TESTRESULTS testRes;


    memset(&vers, 0, sizeof(PACKET70_VERSION));
    memset(&testRes, 0, sizeof (PACKET43_TESTRESULTS));
    memset(message, 0, MESSAGE_MAX);

    message[0] = 0x1b;  // Get the version


    if (ERROR_SUCCESS == NVSOpenPort())
    {
        // Create the message
        dwRetVal = BINRWrite(message, 1, FALSE);

        if (WRITE_OK != dwRetVal)
        {
            RETAILMSG(1, (TEXT("BINRWrite failed with %d\r\n"), dwRetVal));
        }

        Sleep(100);
        dwMsgLen = sizeof(PACKET70_VERSION);
        dwRetVal = BINRRead((BYTE *)&vers, &dwMsgLen);

        if (READ_OK != dwRetVal)
        {
            RETAILMSG(1, (TEXT("BINRRead failed with %d\r\n"), dwRetVal));
        }
        RETAILMSG(1, (TEXT("NVSPort: Version = %S\r\n"),vers.EquipmentSoftwareVersionID));
        RETAILMSG(1, (TEXT("NVSPort: Number of channels = %d\r\n"),vers.numChan));
        RETAILMSG(1, (TEXT("NVSPort: Serial Number = 0x%x/%d\r\n"),vers.SerialNumber,vers.SerialNumber));

        // Create next message to get test results
        message[0] = 0x11;
        dwRetVal = BINRWrite(message, 2, FALSE);
        if (WRITE_OK != dwRetVal)
        {
            RETAILMSG(1, (TEXT("BINRWrite failed with %d\r\n"), dwRetVal));
        }

        Sleep(100);
        dwMsgLen = sizeof (PACKET43_TESTRESULTS);
        dwRetVal = BINRRead((BYTE *)&testRes, &dwMsgLen);

        if (READ_OK != dwRetVal)
        {
            RETAILMSG(1, (TEXT("BINRRead failed with %d\r\n"), dwRetVal));
        }
        RETAILMSG(1, (TEXT("NVSPort: Test Results: 0x%x\r\n"), testRes.TestResults));

        NVSClosePort();
    }
	return 0;
}

